#-------------------------------------------------------------------------------
# Name:        Air Defender
# Purpose:     Basic game
# Author:      PMP
# Created:     18/01/2014
#-------------------------------------------------------------------------------


import pygame as pg
from pygame.locals import *
import sys
import math
import random
from _PyGameColoursPMP import *


def main():
    pg.mixer.pre_init(44100, -16, 2, 2048)  # Set up mixer to avoid sound lag
    pg.init()
    FPS = 30
    fpsClock = pg.time.Clock()
    screenW = 800  # Width of game screen
    screenH = 600  # Height of game screen
    screen = pg.display.set_mode((screenW, screenH))
    pg.display.set_caption('Air Defender PMP')
    gameLoop(screen, screenW, screenH, fpsClock, FPS)


class BackgroundImage():
    surface = 0
    image = None
    imageW = 0
    imageH = 0
    scrollSpeed = 0
    x_pos = 0
    y_pos = 0

    def bgScroll(self, direction):  # N.B. True = 1 (not -1), False = 0
        self.surface.blit(self.image, (self.x_pos, self.y_pos))
        self.surface.blit(self.image,
                          (self.x_pos - self.imageW * (self.x_pos > 0) * (direction == 'R' or direction == 'L'),
                           self.y_pos - self.imageH * (self.y_pos > 0) * (direction == 'D' or direction == 'U')))
        if direction == 'R':
            self.x_pos += self.scrollSpeed
            if self.x_pos >= self.imageW:
                self.x_pos = 0
        elif direction == 'L':
            self.x_pos -= self.scrollSpeed
            if self.x_pos <= 0:
                self.x_pos = self.imageW
        elif direction == 'D':
            self.y_pos += self.scrollSpeed
            if self.y_pos >= self.imageH:
                self.y_pos = 0
        elif direction == 'U':
            self.y_pos -= self.scrollSpeed
            if self.y_pos <= 0:
                self.y_pos = self.imageH


class Player(pg.sprite.Sprite):
    # Variables used by all instances of this class -
    # avoids loading images and sounds each time an instance is created.
    image = pg.image.load('images/SR71.gif')
    image.set_colorkey(image.get_at((0,0)))
    pg.mixer.init(44100, -16, 2, 2048)
    fireSnd = pg.mixer.Sound('sounds/PlayerFire01.wav')

    def __init__(self):
        pg.sprite.Sprite.__init__(self)
        self.image = Player.image.convert()
        self.rect = self.image.get_rect()
        self.fireSnd = Player.fireSnd
        screen = pg.display.get_surface()
        x_topleft, y_topleft, screenW, screenH = screen.get_rect()
        self.step = [8, 8]  # 0 = x_step, 1 = y_step
        imageW = ((self.image.get_width() + self.step[0]) // self.step[0]) * self.step[0]
        imageH = ((self.image.get_height() + self.step[1]) // self.step[1]) * self.step[1]
        self.position = [screenW // 2, screenH - (imageH // 2)]  # 0 = x_pos, 1 = y_pos
        self.bounds = (imageW // 2, 200 + (imageH // 2), screenW - (imageW // 2), screenH - (imageH // 2))
        self.rect.center = self.position  # Ensures image does not temporarily display at 0,0.

    def move(self):  # \ allows code to spread over several lines
        pressed = pg.key.get_pressed()
        if (pressed[pg.K_RIGHT] or pressed[pg.K_x]) \
          and self.position[0] <= self.bounds[2] - self.step[0]:
            self.position[0] += self.step[0]
        elif (pressed[pg.K_LEFT] or pressed[pg.K_z]) \
          and self.position[0] >= self.bounds[0] + self.step[0]:
            self.position[0] -= self.step[0]
        elif (pressed[pg.K_DOWN] or pressed[pg.K_PERIOD]) \
          and self.position[1] <= self.bounds[3] - self.step[1]:
            self.position[1] += self.step[1]
        elif (pressed[pg.K_UP] or pressed[pg.K_SEMICOLON]) \
          and self.position[1] >= self.bounds[1] + self.step[1]:
            self.position[1] -= self.step[1]
        self.rect.center = self.position

    def fire(self, fired, missile_grp):
        pressed = pg.key.get_pressed()
        if (pressed[pg.K_RCTRL] or pressed[pg.K_QUOTE]) and fired == False:
            self.fireSnd.play()
            missile = Missile()
            missile.rect.x = self.rect.x + self.image.get_width() // 2 - 2
            missile.rect.y = self.rect.y
            missile_grp.add(missile)
            fired = True
        return fired, missile_grp


class NonPlayer(pg.sprite.Sprite):
    image = pg.image.load('images/Mig41.gif')
    image.set_colorkey(image.get_at((0,0)))
    image = pg.transform.flip(image, False, True)
    pg.mixer.init(44100, -16, 2, 2048)

    def __init__(self):
        pg.sprite.Sprite.__init__(self)
        self.image = NonPlayer.image.convert()
        self.rect = self.image.get_rect()
        screen = pg.display.get_surface()
        x_topleft, y_topleft, self.screenW, self.screenH = screen.get_rect()
        imageW = self.image.get_width()
        imageH = self.image.get_height()
        x_pos = random.randint(100, self.screenW-100)
        #x_step = random.randint(0, 8) * int(math.copysign(1, random.randint(1, 2) - 2))
        x_step = random.randint(0, 8)
        if x_pos > (self.screenW / 2):
            x_step = -x_step
        y_pos = random.randint(-500, -100)
        y_step = random.randint(1, 8)
        self.position = [x_pos, y_pos]
        self.step = [x_step, y_step]
        self.bounds = [-imageW, -500, self.screenW + imageW, self.screenH + imageH]
        self.rect.center = self.position  # Ensures image does not temporarily display at 0,0.

    def move(self):
        x_pos, y_pos = self.position
        x_step, y_step = self.step
        x_pos += (x_step * (y_pos > -20))  # l-r movement starts only just before plane appears at top of screen.
        y_pos += y_step
        self.position = [x_pos, y_pos]
        self.rect.center = self.position
        alive = True
        if (self.rect.right < 0 or self.rect.left > self.screenW or
            self.rect.top > self.screenH):
                if self.alive() == True:
                    alive = False
                self.kill()
        return alive


class Missile(pg.sprite.Sprite):
    image = pg.Surface([4, 10])
    image.fill(BLACK)
    rect = image.get_rect()
    y_step = 10

    def __init__(self):
        pg.sprite.Sprite.__init__(self)
        screen = pg.display.get_surface()
        x_topleft, y_topleft, self.screenW, self.screenH = screen.get_rect()

    def move(self, fired, missile_grp):
        if fired == True and self.rect.y < self.screenH:
            self.rect.y -= self.y_step
            if self.rect.y < -(self.image.get_height()):
                self.kill()
                missile_grp.empty()
                fired = False
        return fired, missile_grp


def bgScrollSettings(bg, screen, screenW, screenH):
    bgTiled = pg.Surface((screenW, screenH))            # Set up a blank surface the same size as the screen.
    tile = pg.image.load('images/sky01.gif').convert()  # Screen dimensions must be an exact multiple of tile dimensions.
    tileW = tile.get_width()
    tileH = tile.get_height()
    for i in range(screenW // tileW):                   # Fill the blank surface with the tile image.
        for j in range(screenH // tileH):
            bgTiled.blit(tile, (i*tileW, j*tileH))
    bg.image = bgTiled
    bg.imageW = bg.image.get_width()                    # Here bg.imageW = screenW.
    bg.imageH = bg.image.get_height()                   # Here bg.imageH = screenH.
    bg.surface = screen
    bg.scrollSpeed = 2
    bg.x_pos = 0
    bg.y_pos = 0


def displayScore(screen, score, level, lives, timeGone):
    fontObj = pg.font.Font('fonts/freesansbold.ttf', 16)
    #fontObj = pg.font.SysFont('arial.ttf', 16)
    scoretext = fontObj.render('Score : ' + str(score) , True, RED,)
    screen.blit(scoretext, [10, 570])
    leveltext = fontObj.render('Level : ' + str(level) , True, BLUE_D,)
    screen.blit(leveltext, [10, 540])
    livestext = fontObj.render('Lives : ' + str(lives) , True, RED,)
    screen.blit(livestext, [700, 570])
    timetext = fontObj.render('Time : ' + str(timeGone) , True, GREEN,)
    screen.blit(timetext, [10, 10])


def displayCountdown(screen, countdown):
    fontObj = pg.font.Font('fonts/freesansbold.ttf', 64)
    count = fontObj.render(str(countdown), True, RED,)
    count_rect = count.get_rect()
    count_rect.center = screen.get_rect().center
    screen.blit(count, count_rect)


def checkQuit(replay):
    quitGame = False
    for event in pg.event.get():
        if event.type == QUIT or (event.type == KEYUP and event.key == K_ESCAPE):
            quitGame = True
    if quitGame == True or replay == False:
        pg.quit()
        sys.exit('Quitting program - ignore traceback')


#def endGame(screen, playerPlane_grp, enemyPlane_grp, missile_grp):
def endGame(screen):
    fontObj = pg.font.Font('fonts/freesansbold.ttf', 64)
    over = fontObj.render('Game Over', True, RED,)
    over_rect = over.get_rect()
    over_rect.center = screen.get_rect().center
    screen.blit(over, over_rect)
    pg.display.update()
    pg.time.wait(3000)
    fontObj = pg.font.Font('fonts/freesansbold.ttf', 32)
    over = fontObj.render('Press P to play again or Q to quit.', True, GREEN,)
    over_rect = over.get_rect()
    over_rect.centerx = screen.get_rect().centerx
    over_rect.centery = screen.get_rect().centery + 70
    screen.blit(over, over_rect)
    pg.display.update()
    done = False
    while not done:
        for event in pg.event.get():
            if event.type == KEYDOWN:
                if (event.key == K_p):
                    replay = True
                    done = True
                elif (event.key == K_q):
                    replay = False
                    done = True
    return replay


def gameLoop(screen, screenW, screenH, fpsClock, FPS):
    enemyFireSnd = pg.mixer.Sound('sounds/EnemyFire01.wav')
    explosion1Snd = pg.mixer.Sound('sounds/Explode01.wav')
    explosion2Snd = pg.mixer.Sound('sounds/Explode02.wav')
    bgmusicSnd = pg.mixer.music.load('sounds/beethoven-virus.mid')
    pg.mixer.music.play(-1, 0.0)

    explosionImg = pg.image.load('images/Explosion01.gif').convert()

    bg = BackgroundImage()
    bgScrollSettings(bg, screen, screenW, screenH)

    replay = True

    while replay == True:

        playerPlane = Player()
        playerPlane_grp = pg.sprite.GroupSingle(playerPlane)

        maxEnemies = 50
        enemyPlane = [None] * maxEnemies
        for i in range(maxEnemies):
            enemyPlane[i] = NonPlayer()
        enemyPlane_grp = pg.sprite.Group(enemyPlane)

        missile = Missile()
        missile_grp = pg.sprite.GroupSingle(missile)
        missile.kill()
        missile_grp.empty()

        score = 0
        level = 1
        lives = 5
        fired = False

        startTime = pg.time.get_ticks()
        startDelay = 5

        while lives > 0:
            bg.bgScroll('D')
            playerPlane.move()
            fired, missile_grp = playerPlane.fire(fired, missile_grp)
            playerPlane_grp.draw(screen)
            fired, missile_grp = missile.move(fired, missile_grp)
            timeGone = (pg.time.get_ticks() - startTime) // 1000
            if timeGone > startDelay:
                activeEnemies = timeGone // 5
                if activeEnemies > maxEnemies:
                    activeEnemies = maxEnemies
                for i in range(activeEnemies):
                    if enemyPlane[i].move() == False:
                        score -= 1
                        enemyPlane[i] = NonPlayer()        # Creates new enemy to replace killed one.
                        enemyPlane_grp.add(enemyPlane[i])  # Adds the new enemy to the list of enemies.
                    enemyKilled = False
                    #if pg.sprite.collide_rect(playerPlane, enemyPlane[i]):
                    if pg.sprite.spritecollide(enemyPlane[i], playerPlane_grp, False):
                        enemyKilled = True
                        explosion2Snd.play()
                        pg.time.wait(1000)
                        lives -= 1
                        if lives == 0:
                            displayScore(screen, score, level, lives, timeGone)
                            #replay = endGame(screen, playerPlane_grp, enemyPlane_grp, missile_grp)
                            replay = endGame(screen)
                    if pg.sprite.spritecollide(enemyPlane[i], missile_grp, True):
                        enemyKilled = True
                        explosion1Snd.play()
                        score += 10
                        fired = False
                    if enemyKilled == True:
                        enemyPlane[i].kill()
                        enemyPlane[i] = NonPlayer()
                        enemyPlane_grp.add(enemyPlane[i])
            else:
                displayCountdown(screen, startDelay-timeGone)
            enemyPlane_grp.draw(screen)
            missile_grp.draw(screen)
            displayScore(screen, score, level, lives, timeGone-startDelay)
            checkQuit(replay)
            pg.display.update()
            fpsClock.tick(FPS)
            #if replay == True and lives == 0:
            #    break


if __name__ == '__main__':
    main()
